# -*- coding: utf-8 -*-
'''
Input - Output objects for finite state machines.

@date: March 26, 2018
@authors: Damiano Bortolato, Davide Marcato
@email: damiano.bortolato@lnl.infn.it - davide.marcato@lnl.infn.it
'''

import re
from collections import OrderedDict, deque
from statistics import mean, stdev
from math import sqrt
from datetime import datetime
import threading
import epics
from typing import Union, TYPE_CHECKING

# avoid circular import
if TYPE_CHECKING:
    from .fsm import fsmBase


class epicsIO():
    '''Class representing an IO with Epics support for a finite state machine.'''

    def __init__(self, name:str) -> None:
        self._name = name
        self._data = {}  # keep all infos arriving with change callback
        self._conn = False  # keeps all infos arriving with connection callback

        self._attached = set()  # set of finite state machines using this IO
        self._cond = threading.Condition()
        self._pv = epics.PV(name, callback=self.chgcb, connection_callback=self.concb, auto_monitor=True)

    def ioname(self) -> str:
        '''Return the name of the PV.'''
        return self._name

    def attach(self, fsm: 'fsmBase') -> None:
        '''
        Attach a finite state machine to this IO.
        When a FSM is attached to an IO, it will be triggered when the IO changes.
        '''
        self._attached.add(fsm)

    def isAttached(self, fsm: 'fsmBase') -> bool:
        '''Return True if the finite state machine is attached to this IO.'''
        return fsm in self._attached

    def lock(self) -> None:
        '''Obtain exclusive access on IO.'''
        self._cond.acquire()

    def unlock(self) -> None:
        '''Release exclusive access on IO.'''
        self._cond.release()

    def concb(self, **args) -> None:
        '''Callback called on connections and disconnections.'''
        self.lock()
        self._conn = args.get('conn', False)
        self._data = {}  # not to keep old values after disconnection
        self.trigger("conn", args)
        self.unlock()

    def chgcb(self, **args) -> None:
        '''Callback called on value changes, or initial value after connection.'''
        self.lock()
        self._data = args
        self.trigger("change", args)
        self.unlock()

    def putcb(self, **args) -> None:
        '''Callback called when a put() and relative PV processing has completed.'''
        if 'fsm' in args:
            args['fsm'].trigger(iobj=self, inputname=self._name, reason="putcomp")

    def trigger(self, cbname: str, cbdata: dict) -> None:
        '''
        Wakes up all the finite state machines attached to this IO.
        The finite state machines will be triggered with the given callback name and data.
        '''
        for fsm in self._attached:
            fsm.trigger(iobj=self, inputname=self._name, reason=cbname, cbdata=cbdata)

    def put(self, value, caller_fsm: 'fsmBase', use_complete=True) -> bool:
        '''
        Put a value to the PV without wait for the PV processing to complete
        if use_complete=True or issue regular PV put otherwise.
        Returns False if the put() fails to start, may still fail later.
        '''
        # cbdata contains the fsm obj to wake up when putCompleted
        try:
            if use_complete:
                cbdata = {"fsm": caller_fsm}
                self._pv.put(value, callback=self.putcb, use_complete=True, callback_data=cbdata)
            else:
                self._pv.put(value)
        except Exception as e:
            caller_fsm.logE("FAILED putting to pv %s  -- exception: %s" % (self._name, str(e)))
            return False
        return True

    def putComplete(self) -> bool:
        '''Return True if the most recent put() has completed.'''
        return self._pv.put_complete

    def data(self) -> dict:
        '''Return the data dictionary of the PV.'''
        return self._data

    def connected(self) -> bool:
        '''Return True if the PV is connected.'''
        return self._conn


class fsmIOs():
    '''Class representing a list of epicsIO objects.'''

    def __init__(self) -> None:
        self._ios = {}

    def get(self, name: str, fsm: 'fsmBase', **args) -> epicsIO:
        '''Returns the epicsIO object for the given name and attaches it to the given fsm.'''
        
        # first time this input was requested: we create and attach it
        if name not in self._ios:
            self._ios[name] = epicsIO(name)

        # input already created: if not already attached to the fsm, trigger some 
        # fake events to init fsm
        if not self._ios[name].isAttached(fsm):
            io = self._ios[name]
            io.lock()
            fsm.trigger(iobj=io, inputname=name, reason="conn", cbdata={'conn': io.connected(), 'pvname': name})
            if io.data():
                fsm.trigger(iobj=io, inputname=name, reason="change", cbdata=io.data())
            io.attach(fsm)
            io.unlock()

        fsm.logI("Connecting to PV: {:s}".format(name))
        return self._ios[name]

    def getFsmIO(self, fsm: 'fsmBase') -> dict:
        '''Returns a dictionary of all the epicsIO objects attached to the given fsm.'''
        ret = {}
        for io in self._ios.values():
            if io.isAttached(fsm):
                ret[io.ioname()] = io
        return ret

    def getAll(self) -> list:
        '''Returns a list of all the epicsIO objects.'''
        return self._ios.values()


class mappedIOs(fsmIOs):
    '''
    Performs the conversion from procedure internal namings of the inputs
    and real pv names, base on naming convention and a map
    '''
    
    def __init__(self, mapFile) -> None:
        super(mappedIOs, self).__init__()
        # converts the internal name to the ending of the pv name

        file = open(mapFile, "r")
        lines = file.readlines()
        file.close()

        self._map = {}
        replaces = {}  # a dict with macro substitutions
        pattern = OrderedDict()
        strgen = ""
        for line in lines:
            if not line.startswith("#"):
                line_uncomment = line.split("#")[0].strip()
                if line_uncomment.startswith(">"):
                    el = line_uncomment[1:].split("=")
                    if len(el) == 2:
                        cmd = el[0].strip().replace("\"", "")
                        expression = el[1].strip().replace("\"", "")
                        if cmd == "pattern":  # keywords of the config file
                            pattern = OrderedDict()  # will contain the naming convention elements, with each its format
                            strgen = ""  # will contain the whole string formats
                            m = re.match(r" *\((.*)\) *\((.*)\) *", expression)
                            if m:
                                strgen = m.group(1).strip().replace(" ", "").replace("\"", "")  # first part between {} = string definition
                                strelm = re.findall(r"[^{}]*{([^{}]*)}[^{}]*", strgen)  # parse all the single parts
                                patterns = m.group(2).split(",")  # second part between {} = naming convention elements
                                if not strelm or len(strelm) != len(patterns):
                                    raise ValueError("inputMap ERROR, line {}: Failed to parse pattern elements".format(lines.index(line)))
                                for k, kpattern in enumerate(patterns):
                                    p = kpattern.strip().replace(" ", "").replace("\"", "")
                                    if p != "":
                                        if p not in pattern:
                                            pattern[p] = strelm[k]  # populate pattern
                                        else:
                                            raise ValueError("inputMap ERROR, line {}: Redeclaration of pattern element".format(lines.index(line)))
                                if len(pattern) == 0:
                                    raise ValueError("inputMap ERROR, line {}: Pattern empty")
                            else:
                                raise ValueError("inputMap ERROR, line {}: Pattern syntax error".format(lines.index(line)))
                        else:  # macro definitions
                            replaces[el[0].strip().replace("\"", "").upper()] = el[1].strip().replace("\"", "")
                    elif len(el) > 2:
                        raise ValueError("inputMap ERROR, line {}: Multiple or no assignations in line".format(lines.index(line)))
                elif len(line_uncomment) > 3:  # input definition
                    if len(pattern) != 0 and len(strgen) != 0:  # if a pattern has already been defined
                        el = line_uncomment.split("=")
                        if len(el) == 2:  # if there is an assignation
                            key = el[0].strip().replace("\"", "")  # input name inside fsm
                            values = el[1].split(",")  # info to create pv name elements
                            if len(values) != len(pattern):  # must respect the pattern
                                raise ValueError("inputMap ERROR, line {}: Pattern lenght differs from input declaration".format(lines.index(line)))
                            cmap = OrderedDict()  # map of this input: for each element of pattern there is a value
                            for k in range(len(pattern)):
                                candidate = values[k].strip().replace("\"", "")  # parse from file
                                m = re.match(r" *\$\((.*)\) *", candidate)  # $(MACRO)
                                if m:  # if this is a macro to be replaced
                                    candidate = m.group(1)  # get the macro name
                                    if candidate.upper() in replaces:  # if there is a replacement for the macro in replaces
                                        candidate = replaces[candidate.upper()]  # use the replacement
                                    else:
                                        raise ValueError("inputMap ERROR, line {}: Cannot find macro substitutions for: {}".format(lines.index(line), candidate))
                                cmap[list(pattern.items())[k]] = candidate  # the map of this input has a tuple as key (pattern, strelm) and the parsed candidate as value
                            self._map[key] = (cmap, strgen)  # add the current map (of this input) and the whole strgen to the general map (all inputs)
                        else:
                            ValueError("inputMap ERROR, line {}: Multiple or no assignations in line".format(lines.index(line)))
                    else:
                        raise ValueError("inputMap ERROR, line {}: Declaring an input without first defining a pattern!".format(lines.index(line)))

        # inverse map, to perform back naming transformation
        #self.inv_map = {v: k for k, v in self._map.items()}

    # call parent method to connect pvs with complete names
    # reads from calling fsm the targets and creates base pv name with those infos
    def get(self, name: str, fsm: 'fsmBase', **args) -> epicsIO:
        cmap, strgen = self._map[name]

        substitutions = ()  # a tuple containing the parts of pv name in order
        cstrgen = strgen  # copy the string containing the format of each part
        for pattern, v in cmap.items():
            m = re.match(r" *<(.*)> *", v)  # these are parameters to be passed runtime
            if m:
                v = m.group(1)
                if v in args and args[v] is not None:
                    v = args[v]  # get the value from args
                else:
                    raise ValueError("Cannot find the arg: %s in the input creation, as required by input map" % v)
            keydef = pattern[1]
            if keydef.endswith("d"):  # if we are dealing with an int value, let's format it now and then use it as a string
                if v != "":  # this is done so that if we get a number as string doesn't crash
                    v = ("{"+keydef+"}").format(int(v))  # now the number is a string with the correct format
                cstrgen = cstrgen.replace(keydef, ":s", 1)  # so we change the expected format to %s (replace the first occurence, going from first to last)
            elif keydef == ":c":  # this is a char, let's convert it to a string of lenght 1
                v = ("{:.1s}").format(str(v))
                cstrgen = cstrgen.replace(keydef, ":s", 1)
            substitutions += (v,)  # add the updated v to the tuple pv name parts

        pvname = cstrgen.format(*substitutions)  # actually compose pv name
        return super(mappedIOs, self).get(pvname, fsm, **args)

    # return a dictionary with the orinal (before mapping) names of the ios and ios objs of one fsm
    # def getFsmIO(self, fsm):
    #    iosDict = super(lnlPVs, self).getFsmIO(fsm)
    #    pvsDict = {}
    #    for key, value in iosDict.iteritems():
    #        pvsDict[self.inv_map[key]] = value
    #    return iosDict


class fsmIO():
    '''
    An io which changes only between evaluation of the fsm, due to progressive effect of the event 
    queque it reflects the changes of an fsmIO, one change per cycle
    it implements flags to detect changes, edges, connections and disconnections
    there should be a mirror of the same fsmIO for each fsm, in order to use flags indipendently
    '''

    def __init__(self, fsm: 'fsmBase', io: epicsIO) -> None:
        self._fsm = fsm

        self._name = None
        self._value = None  # pv value
        self._sevr = None
        self._pval = None  # pv previous value
        self._psevr = None
        self._timestamp = 0  # value timestamp
        self._currcb = None  # current callback
        self._putComplete = True  # keep track of put completement

        self._reflectedIO = io  # the io to mirror here
        self._name = io.ioname()
        self._conn = False  # pv connected or not
        self._data = {}  # whole pv data
        self.setBufSize(0)

    def setBufSize(self, s: int) -> None:
        '''
        Set the size of the circular buffer for the valAvg and valStd methods.
        '''
        if s == 0:
            self._cbufVal = None
            self._cbufTime = None
        elif self._cbufVal is not None:
            # new buffer with old data and new maxlen
            self._cbufVal = deque(self._cbufVal, maxlen=s)
            self._cbufTime = deque(self._cbufTime, maxlen=s)
        else:
            self._cbufVal = deque(maxlen=s)
            self._cbufTime = deque(maxlen=s)

    def update(self, reason: str, cbdata: dict) -> None:
        '''
        Update this io with the new data from the callback.
        '''
        if reason == 'change':
            self._currcb = reason
            self._data = cbdata
            self._pval = self._value
            self._psevr = self._sevr
            ptime = self._timestamp
            self._value = self._data.get('value', None)
            self._sevr = self._data.get('severity', None)
            self._timestamp = self._data.get('timestamp', 0)
            if self._value is not None and self._cbufVal is not None:
                self._cbufVal.append(self._value)
                # delta in seconds [float] -> microsecond precision
                # how much previous value lasted
                self._cbufTime.append(self._timestamp-ptime)

        elif reason == 'conn':
            self._currcb = reason
            self._conn = cbdata.get('conn', False)
            # on connection or disconnection reset all previous values of the input
            # in order not to access old values after disconnections
            self._pval = None
            self._psevr = None
            self._value = None
            self._sevr = None
            self._data = {}

        # if a put complete callback arrives, the flag must be set true only if
        # the callback was called due to a put made by this object
        elif reason == 'putcomp':
            self._currcb = reason
            self._putComplete = True
        else:
            self._currcb = ""  # a callback which does not modify this input (eg: putcb for other io)

    def reset(self) -> None:
        '''When the state is changed, the current callback is reset'''
        self._currcb = ""

    def ioname(self) -> str:
        '''Return the name of the io'''
        return self._name

    def put(self, value, use_complete=True) -> bool:
        '''Write a value to the io'''
        self._putComplete = False
        return self._reflectedIO.put(value, self._fsm, use_complete)

    # ----- METHODS THAT CATCH CHANGE ONLY if CHECKED WHEN TRIGGERED BY THE SAME CHANGE ------
    # ----- They return True if the fsm was woken up by this change in this cycle

    def putCompleting(self) -> bool:
        '''The current event is a put complete callback of this IO'''
        return self._currcb == 'putcomp'

    def rising(self) -> bool:
        '''The current event increases the value of this IO'''
        return self._currcb == 'change' and self._pval is not None and self._value > self._pval

    def falling(self) -> bool:
        '''The current event decreases the value of this IO'''
        return self._currcb == 'change' and self._pval is not None and self._value < self._pval

    def alarmIncreasing(self) -> bool:
        '''The current event increases the alarm value of this IO'''
        return self._currcb == 'change' and self._psevr is not None and abs(self._sevr) > abs(self._psevr)

    def alarmDecreasing(self) -> bool:
        '''The current event decreases the alarm value of this IO'''
        return self._currcb == 'change' and self._psevr is not None and abs(self._sevr) < abs(self._psevr)

    def alarmChanging(self) -> bool:
        '''The current event changes the alarm value of this IO'''
        return self._currcb == 'change' and self._psevr is not None and self._sevr != self._psevr

    def changing(self) -> bool:
        '''The current event changes the value of this IO'''
        return self._currcb == 'change' and self._pval is not None

    def disconnecting(self) -> bool:
        '''The current event disconnects this IO'''
        return self._currcb == 'conn' and not self._conn

    def connecting(self) -> bool:
        '''The current event connects this IO'''
        return self._currcb == 'conn' and self._conn

    def initializing(self) -> bool:
        '''The current event brings the first value to this IO after connection'''
        return self._currcb == 'change' and self._pval is None

    # ------METHODS THAT KEEP VAlUE BETWEEN TRIGGERS------

    def putComplete(self) -> bool:
        '''The execution of the last `put` command on this IO has been completed'''
        return self._putComplete

    def initialized(self) -> bool:
        '''This IO has received the first value after connection'''
        return self._conn and self._value is not None

    def connected(self) -> bool:
        '''This IO is connected'''
        return self._conn

    def alarm(self) -> int:
        '''The alarm level of this IO'''
        return self._sevr

    def alarmName(self, short=False) -> str:
        '''The alarm level string of this IO'''
        alarm_levels = {-2: "UNDER THRESHOLD MAJOR ALARM",
                        -1: "UNDER THRESHOLD MINOR ALARM",
                         0: "NO ALARM",
                         1: "OVER THRESHOLD MINOR ALARM",
                         2: "OVER THRESHOLD MAJOR ALARM"}
        if short:
            alarm_levels = {k: " ".join(v.split(' ')[-2:]) for k, v in alarm_levels.items()}
        return alarm_levels.get(self._sevr, 'UNKNOWN ALARM')

    def alarmLimits(self) -> tuple:
        '''The alarm thresholds of this IO'''
        lolo = self._data.get('lower_alarm_limit', None)
        low = self._data.get('lower_warning_limit', None)
        high = self._data.get('upper_warning_limit', None)
        hihi = self._data.get('upper_alarm_limit', None)
        return (lolo, low, high, hihi)

    def val(self, as_string=False) -> Union[float, str]:
        '''Current value of this IO'''
        if as_string:
            return self._data.get('char_value', str(self._value))
        return self._value

    def valAvg(self, timeWeight=False) -> float:
        '''
        Current average value of this IO.
        Use `setBufSize` to set the size of the circular buffer for the average.
        If timeWeight is True, the average is weighted by the time passed at each value.
        '''
        if self._cbufVal is None or len(self._cbufVal) < 2:
            return self._value

        if timeWeight:
            times = list(self._cbufTime)[1:].append(datetime.now().timestamp()-self._timestamp)
            vt = [v * t for v, t in zip(self._cbufVal, times)]
            return sum(vt)/sum(times)

        return mean(self._cbufVal)

    def valStd(self, timeWeight=False) -> float:
        '''
        Current standard deviation of this IO.
        Use `setBufSize` to set the size of the circular buffer for the calculation.
        If timeWeight is True, the standard deviation is weighted by the time passed at each value.
        '''
        if self._cbufVal is None or len(self._cbufVal) < 2:
            return 0

        # https://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Weighted_sample_variance
        # https://www.itl.nist.gov/div898/software/dataplot/refman2/ch2/weightsd.pdf
        if timeWeight:
            m = self.valAvg(timeWeight=True)
            times = list(self._cbufTime)[1:].append(datetime.now().timestamp()-self._timestamp)
            wdiff = [t*((v-m)*(v-m)) for v, t in zip(self._cbufVal, times)]
            n = len(times)
            return sqrt(sum(wdiff)/((n-1)*sum(times)/n))

        return stdev(self._cbufVal)

    def valTrend(self, k=1) -> int:
        '''Return the trend of this IO [0 = flat, 1 = increasing, -1 = decreasing]'''
        if self._cbufVal is None or len(self._cbufVal) < 2:
            return 0
        s = stdev(self._cbufVal)
        d = self._cbufVal[-1] - self._cbufVal[0]
        if d > k*s:
            return 1
        if d < -k*s:
            return -1
        return 0

    def pval(self) -> float:
        '''Previous value of this IO'''
        return self._pval

    def time(self) -> datetime:
        '''Timestamp of the last value of this IO'''
        return datetime.fromtimestamp(self._timestamp)

    def status(self) -> int:
        '''The status of the PV (1 for OK)'''
        return self._data.get('status', None)

    def precision(self) -> int:
        '''The precision of the PV (PREC)'''
        return self._data.get('precision', None)

    def units(self) -> str:
        '''The units of the PV (EGU)'''
        return self._data.get('units', None)

    def readAccess(self) -> bool:
        '''The read access of the PV'''
        return self._data.get('read_access', None)

    def writeAccess(self) -> bool:
        '''The write access of the PV'''
        return self._data.get('write_access', None)

    def enumStrings(self) -> list:
        '''Possible string values of enum PV'''
        return self._data.get('enum_strs', None)
    
    def displayLimits(self) -> tuple:
        '''The display limits of the PV (LOPR, HOPR)'''
        l = self._data.get('lower_disp_limit', None)
        u = self._data.get('upper_disp_limit', None)
        return (l, u)

    def controlLimits(self) -> tuple:
        '''The control limits of the PV (DRVL, DRVH)'''
        l = self._data.get('lower_ctrl_limit', None)
        u = self._data.get('upper_ctrl_limit', None)
        return (l, u)

    def maxLen(self) -> int:
        '''The maximum length of the waveform PV (NELM)'''
        return self._data.get('nelm', None)

    def host(self) -> str:
        '''The host of the IOC publishing this PV (IP:port)'''
        return self._data.get('host', None)

    def caType(self):
        '''The CA type of the PV'''
        return self._data.get('type', None)

    def data(self, key=None):
        '''Return one element from pv data, chosen by key, or all data'''
        if key is not None:
            return self._data.get(key, None)
        return self._data



# -*- coding: utf-8 -*-
'''
FSM that implements common operation for other machines.

@date: Jan 20, 2018
@author: Damiamo Bortolato
@email: damiano.bortolato@lnl.infn.it
'''

from . import fsmBase


class fsmTemplate(fsmBase):
    def __init__(self, name, **va):
        fsmBase.__init__(self, name, **va)
        self.setCommonPVs(**va)
        self._waittime = 0
        self._complist = []
        self._retstate = self._curstatename

    def setCommonPVs(self, **va):
        self._errc = va.get('errCodeOut', None)
        self._errm = va.get('errMsgOut', None)
        self._stat = va.get('fsmStateOut', None)
        self._errst = va.get('errStateName', 'error')

    def gotoWait(self, tm, complist, nextstate=None):
        self._waittime = tm
        self._complist = complist
        self._retstate = self._curstatename if nextstate is None else nextstate
        self.gotoState('_wcomp')

    def _wcomp_eval(self):
        if all([x.putComplete() for x in self._complist]):
            self.tmrSet('_tmrwait', self._waittime)
            self.gotoState('_wtimer')

    def _wtimer_eval(self):
        if self.tmrExpired('_tmrwait'):
            self.gotoState(self._retstate)

    def setErrorStatus(self, errCod, errMsg):
        if self._errc:
            self._errc.put(errCod)
        if self._errm:
            self._errm.put(errMsg)

    def gotoState(self, state, *args, **kwargs):
        if self._stat:
            self._stat.put(state)
        return fsmBase.gotoState(self, state, *args, **kwargs)

    def gotoError(self, errCod, errMsg='default error message'):
        self.logE("Error: %s" % errMsg)
        self.setErrorStatus(errCod, errMsg)
        self.gotoState(self._errst)



# -*- coding: utf-8 -*-
'''
The base class of a fsm, which handles stream of execution and provides
user interface to access all the functionalities of io, timers, logger etc.

@date: September 15, 2016
@authors: Damiamo Bortolato, Davide Marcato
@email: damiano.bortolato@lnl.infn.it - davide.marcato@lnl.infn.it
'''

import threading
from .timer import fsmTimers, fsmTimer
from .io import fsmIOs, fsmIO
from .logger import fsmLogger


class fsmBase(threading.Thread):
    '''
    Base class for a finite state machine running in a separate thread
    The user must implement FSM states as methods of the class.
    '''

    def __init__(self, name:str, **args) -> None:
        super(fsmBase, self).__init__(name=name)
        self._name = name
        if 'tmgr' not in args:
            self._private_tmgr = True
            self._tmgr = fsmTimers()
            self._tmgr.start()
        else:
            self._private_tmgr = False
            self._tmgr = args['tmgr']
            if not self._tmgr.is_alive():
                self._tmgr.start()

        self._timers = {}
        self._ios = args.get('ios', fsmIOs())
        self._logger = args.get('logger', fsmLogger())
        self._curstatename = 'undefined'
        self._nextstatename = 'undefined'
        self._prevstatename = 'undefined'
        self._curstate = None
        self._curexit = None
        self._nextstate = None
        self._nextentry = None
        self._nextexit = None
        self._nextstateargs = ((), {})
        self._curstateargs = ((), {})
        self._prevstateargs = ((), {})
        self._senslists = {}
        self._cursens = {}
        self._cond = threading.Condition()
        self._myios = self._ios.getFsmIO(self)
        self._stop_thread = False
        self._events = []
        self._mirrors = {}  # a dict to keep the mirrorIOs of this fsm, with keys the fsmIO
        self._awaker = None  # WHAT FOR FIRST RUN??
        self._awakerType = ""
        self._awakerReason = ""
        self._watchdog = None

    # populate the sensityvity list for each state
    def setSensLists(self, statesWithIos: dict) -> None:
        # statesWithIos e un dizionario in cui la chiave e il nome dello stato e
        # il valore un array di ingressi utilizzati dallo stato
        for state, iolist in statesWithIos.items():
            iodict = {}
            for io in iolist:
                iodict[io.ioname()] = io
            self._senslists[state] = iodict

    def fsmname(self) -> str:
        '''Return the name of this fsm'''
        return self._name

    def lock(self) -> None:
        '''Get exclusive access to this object'''
        self._cond.acquire()

    def unlock(self) -> None:
        '''Release exclusive access to this object'''
        self._cond.release()

    def gotoState(self, state: str, *args, **kwargs) -> None:
        '''Change state to be executed at the next event'''
        self.logD('going to state -> %s' % state)
        if self._nextstate != self._curstate:
            self.logI('gotoState() called twice, ignoring subsequent calls')
            return
        if self._curstatename == state:
            return
        self._nextstatename = state
        self._nextstateargs = (args, kwargs)
        # metodo eval del prossimo stato
        self._nextstate = getattr(self, '%s_eval' % state)
        # metodo entry del prossimo stato
        self._nextentry = getattr(self, '%s_entry' % state, None)
        # metodo exit del prossimo stato
        self._nextexit = getattr(self, '%s_exit' % state, None)

    def gotoPrevState(self, **kwargs) -> None:
        '''Go back to the previous state'''
        if self._prevstatename:
            args = kwargs if kwargs else self._prevstateargs
            self.gotoState(self._prevstatename, **args)

    def log(self, lev: int, msg: str) -> None:
        '''Base method to log a message to the default logger'''
        #whites = len(max(self._senslists.keys(), key=len)) - len(self._curstatename)
        self._logger.log(self._name, lev, '[%15s] : %s' % (self._curstatename, msg))

    def logE(self, msg: str) -> None:
        '''Log an error message'''
        self.log(0, msg)

    def logW(self, msg: str) -> None:
        '''Log a warning message'''
        self.log(1, msg)

    def logI(self, msg: str) -> None:
        '''Log an info message'''
        self.log(2, msg)

    def logD(self, msg: str) -> None:
        '''Log a debug message'''
        self.log(3, msg)

    def logTimeReset(self) -> None:
        self._logger.resetTime()

    def eval(self) -> bool:
        '''Execute the current state'''
        changed = False
        if self._nextstate != self._curstate:
            self.logD('%s => %s' % (self._curstatename, self._nextstatename))
            self._prevstatename = self._curstatename
            self._prevstateargs = self._curstateargs
            self._curstatename = self._nextstatename
            self._curstate = self._nextstate
            self._curstateargs = self._nextstateargs
            self._curexit = self._nextexit
            self._cursens = self._senslists.get(self._nextstatename, {})
            self.commonEntry()
            if self._nextentry:
                self.logD('executing %s_entry()' % (self._curstatename))
                self._nextentry(*self._curstateargs[0], **self._curstateargs[1])
        self.commonEval()
        self.logD('executing %s_eval()' % (self._curstatename))
        self._curstate(*self._curstateargs[0], **self._curstateargs[1])
        self.logD('end of %s_eval()' % (self._curstatename))
        if self._nextstate != self._curstate:
            changed = True
            if self._curexit:
                self.logD('executing %s_exit()' % (self._curstatename))
                self._curexit(*self._curstateargs[0], **self._curstateargs[1])
            self.commonExit()
        return changed

    def eval_forever(self) -> None:
        '''Main loop of the FSM'''
        while not self._stop_thread:
            changed = self.eval()  # eval viene eseguito senza lock
            self.lock()  # blocca la coda degli eventi
            if not changed and len(self._events) == 0:
                self.logD("No events to process going to sleep\n")
                self._cond.wait()  # la macchina va in sleep in attesa di un evento (da un IO, timer...)
                self.logD('awoken')
            self._process_one_event()  # PROCESS ONLY IF RETURN TRUE?
            self.unlock()

    def connect(self, name: str, **args) -> fsmIO:
        '''Create and connect to an fsmIO'''
        thisFsmIO = self._ios.get(name, self, **args)
        if not thisFsmIO in self._mirrors:
            self._mirrors[thisFsmIO] = fsmIO(self, thisFsmIO)

        return self._mirrors[thisFsmIO]

    def run(self) -> None:
        '''Execute the FSM thread'''
        self.eval_forever()

    def kill(self) -> None:
        '''Stop the FSM thread'''
        self._cond.acquire()
        self._stop_thread = True
        self._cond.notify()
        self._cond.release()
        self.join()
        if self._private_tmgr:
            self._tmgr.kill()

    def trigger(self, **args) -> None:
        '''Enqueue an event to be processed by the FSM'''
        self._cond.acquire()
#        self.logD("pushing event %s %d" %(repr(args), len(self._events)+1))
        self._events.append(args)  # append here also the shapshot of all ios
        if len(self._events) == 1:
            self._cond.notify()
        self._cond.release()

    def _process_one_event(self) -> bool:
        '''Process one event from the queue'''
        if self._awaker and self._awakerType == 'io':
            # reset io to catch changes only on the eval triggered by the same change
            self._awaker.reset()
        self.resetAwaker()
        if len(self._events):
            return self._process_event(**self._events.pop(0))
        return False

    def _process_event(self, **args) -> bool:
        '''Apply the updates carried by the event'''
        self.logD('Consuming event n° %d' % (len(self._events)))
        if 'inputname' in args:
            self.logD("input " + repr(args['inputname']) +
                      " is triggering " + self._curstatename +
                      " - " + args['reason'])
            fsmIOobj = args['iobj']
            mirrorIOobj = self._mirrors.get(fsmIOobj, None)
            if mirrorIOobj:
                mirrorIOobj.update(args.get('reason', ""), args.get('cbdata', None))
                self.setAwaker(mirrorIOobj, 'io', args.get('reason', "unknownInput"))
                return True
        elif 'timername' in args:
            self.logD("timer " + repr(args['timername']) + " is triggering " + self._curstatename)
            self.setAwaker(args['tmrobj'], 'tmr', args.get('reason', "unknownTimer"))
            return True
        return False

    def commonEval(self):
        pass

    def commonExit(self):
        pass

    def commonEntry(self):
        pass

    def tmrSet(self, name: str, timeout: float, reset=True) -> None:
        '''Set a timer with a given timeout in seconds'''
        if not name in self._timers:
            self._timers[name] = fsmTimer(self, name)
        t = self._timers[name]
        self._tmgr.set(t, timeout, reset)
        self.logD("activating a timer: '%s', %.2fs" % (name, timeout))

    def tmrExpired(self, name: str) -> bool:
        '''Check if a timer has expired or does not exist'''
        return not name in self._timers or self._timers[name].expd()

    def tmrExpiring(self, name: str) -> bool:
        '''The current event is caused by the expiration of a this timer'''
        return name in self._timers \
            and self._awakerType == 'tmr' \
            and self._awakerReason == "expired" \
            and self._awaker is self._timers[name]

    def isIoConnected(self) -> bool:
        '''Check if all the IOs are connected'''
        stateios = self._cursens if self._cursens else self._mirrors
        return self.allof(stateios.values(), "connected")

    def isIoInitialized(self) -> bool:
        '''Check if all the IOs are initialized'''
        stateios = self._cursens if self._cursens else self._mirrors
        return self.allof(stateios.values(), "initialized")

    def anyof(self, objs: list, method: str) -> bool:
        '''Check if any of the IOs is in a given state'''
        return any(getattr(io, method)() for io in objs)

    def allof(self, objs: list, method: str) -> bool:
        '''Check if all the IOs are in a given state'''
        return all(getattr(io, method)() for io in objs)

    def whoWokeMe(self) -> tuple:
        '''Return the object that woke up the FSM'''
        return (self._awaker, self._awakerType)

    def whyWokeMe(self) -> str:
        '''Return the reason that woke up the FSM'''
        return self._awakerReason

    def setAwaker(self, obj: object, objType: str, reason: str) -> None:
        '''Set the object that woke up the FSM'''
        self._awaker = obj
        self._awakerType = str(objType)
        self._awakerReason = str(reason)

    def resetAwaker(self) -> None:
        '''Reset the object that woke up the FSM'''
        self._awaker = None
        self._awakerType = ""
        self._awakerReason = ""

    def setWatchdogInput(self, inp: fsmIO, mode="on-off", interval=1) -> None:
        if isinstance(inp, fsmIO) and mode in ["off", "on", "on-off"]:
            self._watchdog = (inp, mode, interval)
        else:
            raise ValueError("Unrecognized input type or mode")

    def getWatchdogInput(self) -> fsmIO:
        return self._watchdog



# -*- coding: utf-8 -*-
'''
Timers to awake a fsm after a certain amount of time.

@date: March 26, 2018
@authors: Damiamo Bortolato, Davide Marcato
@email: damiano.bortolato@lnl.infn.it - davide.marcato@lnl.infn.it
'''

import threading
import time
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .fsm import fsmBase


class fsmTimer():
    '''A timer to awake a fsm after a certain amount of time.'''

    def __init__(self, fsm: 'fsmBase', name: str) -> None:
        self.expire = 0
        self._fsm = fsm
        self._pending = False
        self._name = name

    def reset(self, timeout: float) -> None:
        '''Reset the timer to expire after timeout seconds.'''
        self.expire = time.time() + timeout
        self._pending = True

    def trigger(self) -> None:
        '''Trigger the corresponding FSM.'''
        self._pending = False
        self._fsm.trigger(tmrobj=self, timername=self._name, reason="expired")

    def name(self) -> str:
        '''Return the name of the timer.'''
        return self._name

    def expd(self) -> bool:
        '''Return if the timer is expired or not.'''
        return not self._pending


class fsmTimers(threading.Thread):
    '''A thread to trigger timers when they expire.'''

    def __init__(self) -> None:
        threading.Thread.__init__(self)
        # lock per l'accesso esclusivo
        self._cond = threading.Condition()
        # array con i timers attivi in ordine di scadenza
        # (indice 0 è il prossimo a scadere)
        self._timers = []
        self._stop_thread = False

    def run(self) -> None:
        '''
        Main thread routine.
        It sleeps for the time that remains to the next timer to expire
        (the first of a list ordered by expiration time). When the sleep expires, the thread
        starts to see how many timers have expired starting from the next one.
        More than one can expire when they have the same expiration time or the intervals
        fall into the jitter of the thread execution. For each expired timer it executes the
        trigger and removes it from the list of pending timers.
        '''
        # acquisisce il lock, per avere accesso esclusivo alla lista dei timer
        self._cond.acquire()
        next_wait = None
        while not self._stop_thread:
            if len(self._timers):  # se ci sono timers in pendenti
                now = time.time()  # tempo corrente
                i = 0
                next_wait = None
                for t in self._timers:
                    if t.expire > now:
                        # abbiamo trovato il primo timer non ancora scaduto,
                        # interrompo la scansione, i rappresenta l'indice del primo timer non scaduto
                        next_wait = self._timers[i].expire - now
                        break
                    i += 1
                # triggera gli eventi per i timer che vanno da 0 a i-1
                for t in self._timers[:i]:
                    t.trigger()
                # rimuove i primi 'i' timers (che sono scaduti)
                self._timers = self._timers[i:]
            # va in sleep per i prossimi 'next_wait' secondi, ovvero l'intervallo di tempo al termine del quale scadra'
            # il prossimo timer. Se non ci sono timer va in sleep a per un tempo indefinito
            # NB: wait rilascia il lock permettendo ad altri thread di impostare altri timer
            self._cond.wait(next_wait)

    def set(self, timer: fsmTimer, timeout: float, reset=True) -> None:
        '''Set a fsmTimer to expire after timeout seconds.'''

        # ottiene l'accesso esclusivo alla lista dei timer
        self._cond.acquire()
        try:
            # se il timer è già in lista significa che è stato reimpostato prima che scadesse,
            # quindi lo rimuovo e lo reimposto
            if timer in self._timers:
                if not reset:
                    # timer già settato e non è richiesto reset, ritorno (il release() della
                    # condizione è in finally:
                    return
                self._timers.remove(timer)

            # imposta il tempo al quale scadrà il timer
            timer.reset(timeout)

            i = 0
            for t in self._timers:
                if t.expire > timer.expire:
                    # il timer all'indice 'i' scade dopo il timer che sto impostando, pertanto
                    # inserisco il nuovo timer in questa posizione 'i' e interrompo il ciclo
                    break
                i += 1
            self._timers.insert(i, timer)
            if i == 0:
                # CASO SPECIALE: se 'i'  == 0 significa che ho inserito in testa il nuovo timer oppure l'ho inserito
                # in una lista vuota; nel primo caso devo svegliare il thread perche' il nuovo timer scadra' prima
                # del suo prossimo risveglio (impostato su quello che ora è il secondo timer in lista), nel secondo
                # caso il thread sta dormendo per un tempo indefinito, quindi lo devo svegliare affinche reimposti
                # un tempo di sleep corretto
                self._cond.notify()
            # rilascia il lock
        except Exception as e:
            print(repr(e))
        finally:
            self._cond.release()

    def kill(self):
        '''Kill the thread.'''
        self._cond.acquire()
        self._stop_thread = True
        self._cond.notify()
        self._cond.release()
        self.join()



from smlib import fsmBase, loader
import pytest
from queue import Queue, Empty
from pcaspy import SimpleServer, Driver
from pcaspy.tools import ServerThread
import epics

# Test IO
# - Check main methods of epics IO


# Dummy PV driver, just store PV values as set by external puts
class ioDriver(Driver):
    def __init__(self):
        super().__init__()


class FSM(fsmBase):
    def __init__(self, name, **args):
        super().__init__(name, **args)
        self.events = Queue(maxsize=1000)
        self.exec_n = 0

        self.pv1 = self.connect('Ts:pv1')
        #self.pv2 = self.connect('pv2')
        #self.wf = self.connect('wf')

        self.gotoState('idle')

    def register_event(self, name):
        ev = {"name": name}
        ev["pv1_connected"] = self.pv1.connected()
        ev["pv1_connecting"] = self.pv1.connecting()
        ev["pv1_disconnecting"] = self.pv1.disconnecting()
        ev["pv1_initialized"] = self.pv1.initialized()
        ev["pv1_initializing"] = self.pv1.initializing()
        ev["pv1_changing"] = self.pv1.changing()
        ev["awaker"] = self.whoWokeMe()[0]
        ev["reason"] = self.whyWokeMe()
        ev["value"] = self.pv1.val()
        ev["exec_n"] = self.exec_n
        print(ev)
        self.events.put(ev)

    def idle_entry(self):
        self.exec_n = 0
        self.register_event("idle_entry")

    def idle_eval(self):
        self.exec_n += 1
        self.register_event("idle_eval")


# Will be called to produce the argument to test_io function
# and again after its execution to kill all the FSMs
@pytest.fixture(scope='module')
def loaded_fsm():
    l = loader()
    fsm = l.load(FSM, "test_io")
    l.start(blocking=False)

    yield fsm
    # After tests have completed or failed
    l.killAll(None, None)


# Will be called to produce the argument to test_io function
# and again after its execution to kill all the FSMs
@pytest.fixture
def pv_server(scope='function'):
    prefix = 'Ts:'
    pvdb = {
        'pv1': {
            'value': 0,
            'hihi': 10,
            'high':  5,
            'low': -5,
            'lolo': -10
        },
        'pv2': {},
        'wf': {
            'type': 'char',
            'count': 300,
            'value': 'some initial message. but it can become very long.'
        }
    }

    server = SimpleServer()
    server.createPV(prefix, pvdb)
    driver = ioDriver()
    server_thread = ServerThread(server)
    server_thread.start()

    yield server_thread

    # After tests have completed or faileds
    server_thread.stop()

# Test function to check timers expected behaviour
def test_connection(loaded_fsm, pv_server):
    events = loaded_fsm.events
    pv1 = epics.PV('Ts:pv1')
    #pv2 = epics.PV('pv2')
    #wf = epics.PV('wf')

    # first automatic run of idle_entry()
    ev = events.get(timeout=1)
    assert ev["name"] == "idle_entry"
    assert ev["awaker"] is None
    assert not ev["pv1_connected"]
    assert not ev["pv1_connecting"]
    assert not ev["pv1_initialized"]
    assert not ev["pv1_initializing"]
    assert not ev["pv1_changing"]
    assert ev["value"] is None
    assert ev["exec_n"] == 0

    # first automatic run of idle_eval()
    ev = events.get(timeout=1)
    assert ev["name"] == "idle_eval"
    assert ev["awaker"] is None
    assert not ev["pv1_connected"]
    assert not ev["pv1_connecting"]
    assert not ev["pv1_initialized"]
    assert not ev["pv1_initializing"]
    assert not ev["pv1_changing"]
    assert ev["value"] is None
    assert ev["exec_n"] == 1

    # fake connection event to initialize fsm
    ev = events.get(timeout=1)
    assert ev["name"] == "idle_eval"
    assert ev["awaker"] is not None
    assert not ev["pv1_connected"]
    assert not ev["pv1_connecting"]
    assert not ev["pv1_initialized"]
    assert not ev["pv1_initializing"]
    assert not ev["pv1_changing"]
    assert ev["value"] is None
    assert ev["exec_n"] == 2

    # connection event
    ev = events.get(timeout=1)
    assert ev["name"] == "idle_eval"
    assert ev["awaker"] is not None
    assert ev["pv1_connected"]
    assert ev["pv1_connecting"]
    assert not ev["pv1_initialized"]
    assert not ev["pv1_initializing"]
    assert not ev["pv1_changing"]
    assert ev["value"] is None
    assert ev["exec_n"] == 3

    # first change event to initialize the IO
    ev = events.get(timeout=1)
    assert ev["name"] == "idle_eval"
    assert ev["awaker"] is not None
    assert ev["pv1_connected"]
    assert not ev["pv1_connecting"]
    assert ev["pv1_initialized"]
    assert ev["pv1_initializing"]
    assert not ev["pv1_changing"]
    assert ev["value"] == 0
    assert ev["exec_n"] == 4

    # change event
    pv1.put(3)
    ev = events.get(timeout=1)
    assert ev["name"] == "idle_eval"
    assert ev["awaker"] is not None
    assert ev["pv1_connected"]
    assert not ev["pv1_connecting"]
    assert ev["pv1_initialized"]
    assert not ev["pv1_initializing"]
    assert ev["pv1_changing"]
    assert ev["value"] == 3
    assert ev["exec_n"] == 5


def test_disconnection(loaded_fsm):
    events = loaded_fsm.events

    # Disconnection event
    ev = events.get(timeout=1)
    assert ev["name"] == "idle_eval"
    assert ev["awaker"] is not None
    assert not ev["pv1_connected"]
    assert not ev["pv1_connecting"]
    assert ev["pv1_disconnecting"]
    assert not ev["pv1_initialized"]
    assert not ev["pv1_initializing"]
    assert not ev["pv1_changing"]
    assert ev["value"] is None
    assert ev["exec_n"] == 6

    # No more events
    with pytest.raises(Empty):
        ev = events.get(timeout=0.5)



import pytest
from smlib import fsmBase, loader, fsmLogger, fsmFileLogger, fsmIOs, mappedIOs

@pytest.fixture
def l(scope='module'):
    return loader()
    

def test_setVerbosity(l):
    #default value 2=info
    assert l._verbosity == 2
    assert l._logger._level == 2 
    
    # test setting verbosity via number
    l.setVerbosity(-1)
    assert l._logger._level == 0
    assert l._verbosity == 0
    l.setVerbosity(0)
    assert l._verbosity == 0
    assert l._logger._level == 0
    l.setVerbosity(1)
    assert l._verbosity == 1
    assert l._logger._level == 1
    l.setVerbosity(2)
    assert l._verbosity == 2
    assert l._logger._level == 2
    l.setVerbosity(3)
    assert l._verbosity == 3
    assert l._logger._level == 3
    l.setVerbosity(4)
    assert l._verbosity == 3
    assert l._logger._level == 3

    # test setting verbosity via string
    l.setVerbosity("error")
    assert l._verbosity == 0
    assert l._logger._level == 0
    l.setVerbosity("warning")
    assert l._verbosity == 1
    assert l._logger._level == 1
    l.setVerbosity("info")
    assert l._verbosity == 2
    assert l._logger._level == 2
    l.setVerbosity("debug")
    assert l._verbosity == 3
    assert l._logger._level == 3
    with pytest.raises(KeyError):
        l.setVerbosity("otherstring")


def test_logToFile(l):
    #default value
    assert isinstance(l._logger, fsmLogger)

    l.logToFile("dir", "prefix")
    assert isinstance(l._logger, fsmFileLogger)
    assert l._logger.dir == "dir"
    assert l._logger.prefix == "prefix"
    assert l._logger._level == l._verbosity


def test_setIoMap(l):
    #default
    assert isinstance(l._ioManager, fsmIOs)

    l.setIoMap("../examples/config/iomap.txt")
    assert isinstance(l._ioManager, mappedIOs)

def test_load(l):
    #default
    assert len(l._fsmsList) == 0

    class fsm_fake(fsmBase):
        def __init__(self, name, num, **kwargs):
            self.timerManager = kwargs.get("tmgr", None)
            self.ioManager = kwargs.get("ios", None)
            self.logger = kwargs.get("logger", None)
            self.args = num
        
    l.load(fsm_fake, "testname", 5)
    assert len(l._fsmsList) == 1
    loaded = l._fsmsList[0]
    assert isinstance(loaded, fsm_fake)
    assert loaded.timerManager == l._timerManager
    assert loaded.ioManager == l._ioManager
    assert loaded.logger == l._logger
    assert loaded.args == 5


#! /usr/bin/python
# -*- coding: utf-8 -*-
'''
Implementing watchdogs for others fsm.

@date: Oct 2016
@author: Davide Marcato
@email: davide.marcato@lnl.infn.it
'''

from random import uniform
from . import fsmBase


class fsmWatchdog(fsmBase):
    '''A FSM that implements watchdogs for others fsm.'''

    def __init__(self, name: str, fsms: list, **args) -> None:
        super(fsmWatchdog, self).__init__(name, **args)

        # a list of timers with timer name and input linked
        self.timers = {}
        # connect input for each fsm
        # the input is the pv where to write each time to say the fsm is alive
        for fsm in fsms:
            wd = fsm.getWatchdogInput()
            if wd is not None:
                inp = wd[0]
                self.timers[inp.ioname()] = fsm

        statesWithIOs = {
            "run": []
        }
        self.setSensLists(statesWithIOs)
        self.gotoState('run')

    def run_entry(self):
        for tmrName, fsm in self.timers.items():
            # start with a random delay not to write all pvs at the same instant
            interval = fsm.getWatchdogInput()[2]
            randDelay = uniform(0, interval)
            self.tmrSet(tmrName, randDelay)
        self.logD("set %d watchdogs" % len(self.timers))

    def run_eval(self):
        for tmrName, fsm in self.timers.items():
            if self.tmrExpired(tmrName):
                inp, mode, interval = fsm.getWatchdogInput()
                self.tmrSet(tmrName, interval)
                if inp.connected():
                    if mode == "on":
                        inp.put(int(fsm.is_alive()))
                    elif mode == "off":
                        inp.put(int(not fsm.is_alive()))
                    elif mode == "on-off":
                        if fsm.is_alive():
                            inp.put(int(not inp.val()))
                    else:
                        self.logE("Unknown watchdog mode")



from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

__doc__ = """
   A python library for creating EPICS finite state machines, 
   running in different threads as daemons and sharing resources.

   version: %s
   Principal Authors:
      Damiano Bortolato <damiano.bortolato@lnl.infn.it> INFN, Laboratori Nazionali di Legnaro
      Davide Marcato <davide.marcato@lnl.infn.it> INFN, Laboratori Nazionali di Legnaro

== License:
   GPLv3 - This is free software; see the source for copying conditions.  There is NO
   warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

== Overview:
   Main features include:
    - Easy to use and fast deployment of complex event based fsm.
    - High expandability as provided by all the libraries of Python.
    - Integrated configurables logging systems.
    - Convenient methods to access all the information on I/O.
    - Timers can be used to execute actions after a time delay.
    - Integrated watchdog logic.
    - Multi-threading: each fsm is executed on a different thread, sharing I/O.
    - Convenient loader to launch a daemon with multiple fsm.
    - Possibility to apply a configurable naming convention on I/O.

""" % (__version__)

from .fsm import fsmBase
from .logger import fsmLogger, fsmFileLogger
from .timer import fsmTimers
from .io import fsmIOs, mappedIOs, fsmIO
from .fsmTemplate import fsmTemplate
from .fsmWatchdog import fsmWatchdog
from .loader import loader



#! /usr/bin/python
'''
A loader to simplify the creation of a daemon of various fsm running
on different threads and sharing resources.

@date: February 2018
@authors: Davide Marcato
@email: davide.marcato@lnl.infn.it
'''

import signal
from . import fsmFileLogger, fsmLogger, mappedIOs, fsmIOs, fsmTimers, fsmWatchdog, fsmBase


# class to load multiple fsm
class loader(object):
    '''A class to load multiple fsm and manage them.'''

    def __init__(self) -> None:
        self._timerManager = fsmTimers()
        self._verbosity = 2
        self._logger = fsmLogger(self._verbosity)
        self._ioManager = fsmIOs()
        self._ioMap = None
        self._fsmsList = []
        self._levelStrings = {"error": 0, "warning": 1, "info": 2, "debug": 3}

    def setVerbosity(self, level: int or str) -> None:
        '''Set the verbosity level of the logger.'''
        if isinstance(level, int):
            n = max(0, min(level, 3))  # log level must be in range [0,3]
        elif isinstance(level, str) and level.lower().strip() in self._levelStrings.keys():
            n = self._levelStrings[level.lower().strip()]
        else:
            raise KeyError("Verbosity level \"%s\" not recognized!" % str(level))

        self._verbosity = n
        self._logger.changeLevel(n)

    def logToFile(self, path: str, prefix: str):
        '''Set the logger to log to file.'''
        self._logger = fsmFileLogger(self._verbosity, path, prefix)

    def setIoMap(self, iomap: str):
        '''Set the iomap to use for the IOs.'''
        self._ioMap = iomap
        self._ioManager = mappedIOs(self._ioMap)

    def load(self, fsmClass: type, name: str, *args, **kwargs):
        '''Load a fsm class to be executed.'''
        kwargs["tmgr"] = self._timerManager
        kwargs["ios"] = self._ioManager
        kwargs["logger"] = self._logger
        if not issubclass(fsmClass, fsmBase):
            raise TypeError("%s is not a subclass of fsmBase" % repr(fsmClass))
        fsm = fsmClass(name, *args, **kwargs)  # instance class
        self._fsmsList.append(fsm)
        return fsm

    def killAll(self, signum, frame): # pylint: disable=unused-argument
        '''Kill all the fsms and the timer manager.'''
        #print("Signal: %d -> Going to kill all fsms" % signum)
        for fsm in self._fsmsList:
            if fsm.is_alive():
                fsm.kill()
        print("Killed all the fsms")
        if self._timerManager.is_alive():  # if no fsm is loaded it won't be alive
            self._timerManager.kill()
        print("Killed the timer manager")

    def printUnconnectedIOs(self, signum, frame): # pylint: disable=unused-argument
        '''Print all the unconnected IOs.'''
        ios = self._ioManager.getAll()
        s = 0
        print("DISCONNECTED INPUTS:")
        for i in ios:
            if not i.connected():
                print(i.ioname())
                s += 1
        print("Total disconnected inputs: %d out of %d!" % (s, len(ios)))
        signal.pause()

    def start(self, blocking=True):
        '''Start all the loaded fsms.'''
        # start another fsm to report if all the others are alive to epics db
        wd = fsmWatchdog("REPORT", self._fsmsList, tmgr=self._timerManager, ios=self._ioManager, logger=self._logger)
        self._fsmsList.append(wd)

        for thread in self._fsmsList:
            thread.start()
        print("%d fsms started!" % (len(self._fsmsList)-1))  # do not count fsmWatchdog (not issued by user)

        if blocking:
            # wait for events
            signal.signal(signal.SIGINT, self.killAll)
            signal.signal(signal.SIGUSR1, self.printUnconnectedIOs)
            signal.pause()



'''Test FSM states execution with args'''

from smlib import fsmBase, loader
import pytest
import time

class FSM(fsmBase):
    def __init__(self, name, **args):
        super().__init__(name, **args)
        self.counter = []
        self.gotoState("idle")

    def idle_entry(self):
        self.counter.append(0)

    def idle_eval(self):
        self.counter.append(1)

    def only_args_entry(self, num):
        self.counter.append(num-1)

    def only_args_eval(self, num):
        self.counter.append(num)
        self.gotoState("idle")

    def only_args_exit(self, num):
        self.counter.append(num+1)

    def third_entry(self, num):
        self.counter.append(num-1)

    def only_kwargs_entry(self, num=100):
        self.counter.append(num-1)

    def only_kwargs_eval(self, num=100):
        self.counter.append(num)
        self.gotoState("idle")

    def only_kwargs_exit(self, num=100):
        self.counter.append(num+1)

    def args_and_kwargs_entry(self, num, num2=-100):
        self.counter.append(num-1)
        self.counter.append(num2-1)

    def args_and_kwargs_eval(self, num, num2=-100):
        self.counter.append(num)
        self.counter.append(num2)
        self.gotoState("idle")

    def args_and_kwargs_exit(self, num, num2=-100):
        self.counter.append(num+1)
        self.counter.append(num2+1)


# Will be called to produce the argument to test_io function
# and again after its execution to kill all the FSMs
@pytest.fixture(scope='module')
def loaded_fsm():
    l = loader()
    fsm = l.load(FSM, "test_io")
    l.start(blocking=False)

    yield fsm
    # After tests have completed or failed
    l.killAll(None, None)



def test_args(loaded_fsm):
    '''Test gotoState with args'''
    assert loaded_fsm.counter == [0, 1]
    loaded_fsm.gotoState("only_args", 10)
    loaded_fsm.trigger()
    time.sleep(0.1)
    assert loaded_fsm.counter == [0, 1, 9, 10, 11, 0, 1]

def test_kwargs(loaded_fsm):
    '''Test gotoState with kwargs'''
    loaded_fsm.counter = []
    loaded_fsm.gotoState("only_kwargs")
    loaded_fsm.trigger()
    time.sleep(0.1)
    assert loaded_fsm.counter == [99, 100, 101, 0, 1]

def test_kwargs_filled(loaded_fsm):
    '''Test gotoState with kwargs with value'''
    loaded_fsm.counter = []
    loaded_fsm.gotoState("only_kwargs", num=50)
    loaded_fsm.trigger()
    time.sleep(0.1)
    assert loaded_fsm.counter == [49, 50, 51, 0, 1]

def test_args_kwargs(loaded_fsm):
    '''Test gotoState with args and kwargs unfilled'''
    loaded_fsm.counter = []
    loaded_fsm.gotoState("args_and_kwargs", 10)
    loaded_fsm.trigger()
    time.sleep(0.1)
    assert loaded_fsm.counter == [9, -101, 10, -100, 11, -99, 0, 1]

def test_args_kwargs_filled(loaded_fsm):
    '''Test gotoState with args and kwargs filled'''
    loaded_fsm.counter = []
    loaded_fsm.gotoState("args_and_kwargs", 10, num2=50)
    loaded_fsm.trigger()
    time.sleep(0.1)
    assert loaded_fsm.counter == [9, 49, 10, 50, 11, 51, 0, 1]



# Python installation file
#
# TO INSTALL (user):
# pip install [-e] .
#
# TO UPLOAD NEW RELEASE (maintainer):
# 0 - Update CHANGELOG.md
# 1 - git tag <new_tag>
# 2 - python setup.py clean --all
# 3 - rm -rf dist
# 4 - python setup.py sdist bdist_wheel
# 5 - twine upload -r testpypi dist/*
# 6 - twine upload dist/*
# 7 - cd docs/ && make html
# 8 - git commit -am "Documentation update"
# 9 - git push github master

from setuptools import setup
import versioneer

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='pysmlib',
      version=versioneer.get_version(),
      description='Python Finite State Machines for EPICS',
      python_requires='>=3.6',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://darcato.github.io/pysmlib/docs/html/',
      download_url='https://github.com/darcato/pysmlib',
      author='Damiano Bortolato - Davide Marcato',
      author_email='davide.marcato@lnl.infn.it',
      license='GPLv3',
      packages=['smlib'],
      install_requires=['pyepics'],
      zip_safe=False)



# -*- coding: utf-8 -*-
'''
Providing a way to easily log the behaviour of a fsm.

@date: March 26, 2018
@authors: Damiamo Bortolato, Davide Marcato
@email: damiano.bortolato@lnl.infn.it - davide.marcato@lnl.infn.it
'''

import time
from datetime import datetime


class fsmLogger():
    '''Base class for fsm loggers.'''
    
    levstr = ['E', 'W', 'I', 'D']

    def __init__(self, lev=3) -> None:
        self._level = lev
        self.startime = time.time()

    def log(self, fsmname: str, lev: int, msg: str) -> None:
        tm = time.time() - self.startime
        if lev <= self._level:
            self.pushMsg('%8.2fs: %s - %s%s' % (tm, fsmLogger.levstr[lev], fsmname, msg))

    def pushMsg(self, msg: str) -> None:
        print(msg)

    def resetTime(self) -> None:
        self.startime = time.time()

    def changeLevel(self, newlevel: int) -> None:
        self._level = newlevel


class fsmFileLogger(fsmLogger):
    '''A class to log to file.'''

    def __init__(self, lev=3, directory="logs/", prefix="") -> None:
        super(fsmFileLogger, self).__init__(lev)
        self.files = {}
        self.dir = directory
        self.prefix = prefix

    def log(self, fsmname: str, lev: int, msg: str) -> None:
        '''Log a message to file.'''
        if lev <= self._level:
            if fsmname not in self.files.keys():
                self.files[fsmname] = open(self.dir+"/"+self.prefix+"."+fsmname+".log", 'a')
            tm = datetime.now()
            self.pushMsg(self.files[fsmname], '%s: %s - %s\n' % (str(tm), fsmLogger.levstr[lev], msg))

    def pushMsg(self, f, msg: str):
        '''Write a message to file.'''
        f.write(msg)
        f.flush()  # to empty buffer after every message!

    def __del__(self):
        '''Close all the open files.'''
        for name, f in self.files.items():
            if not f.closed:
                print("Closing "+name+"\n")
                f.close()

# TODO: add fsmSysLogger class



from smlib import fsmBase, loader
from datetime import datetime
import pytest
from queue import Queue, Empty
import time

# Test timers
# - must wake up the current state
# - must do it after the correct time delta
# - before being set must return expired == True
# - while waiting must return expired == False
# - after expiration must return expired == True
# - (only) during the expiration event must return exipiring == True

class FSM(fsmBase):
    def __init__(self, name, **args):
        super().__init__(name, **args)
        self.events = Queue()
        self.t = datetime.now()
        self.exec_n = 0
        self.gotoState('run')
    
    def register_event(self, name):
        ev = {"name": name}
        ev["t1_expired"] = self.tmrExpired('t1')
        ev["t1_expiring"] = self.tmrExpiring('t1')
        ev["delta"] = datetime.now() - self.t
        ev["exec_n"] = self.exec_n
        self.events.put(ev)

    def run_entry(self):
        self.register_event("before_set")
        self.t = datetime.now()
        self.tmrSet('t1', 0.2)
        self.tmrSet('t1', 0.5) # default reset = True, the active timer is reinitialized
        self.register_event("after_set")
        self.tmrSet('t2', 0.7)
        self.tmrSet('t2', 1, reset=False) # reset = False, the active timer is not changed
        self.exec_n = 0

    def run_eval(self):
        self.register_event("run_eval")
        self.exec_n +=1


# Will be called to produce the argument to test_timer function
# and again after its execution to kill all the FSMs
@pytest.fixture
def loaded_fsm(scope='function'):
    l = loader()
    fsm = l.load(FSM, "test_timer")
    l.start(blocking=False)
    yield fsm
    l.killAll(None, None)

# Test function to check timers expected behaviour
def test_timer(loaded_fsm):
    events = loaded_fsm.events
    
    # before timer t1 is set
    ev = events.get(timeout=1)
    assert ev["name"]=="before_set"
    assert ev["t1_expired"]
    assert not ev["t1_expiring"]

    # after timer t1 is set
    ev = events.get(timeout=1)
    assert ev["name"]=="after_set"
    assert not ev["t1_expired"]
    assert not ev["t1_expiring"]

    # direct first run of run_eval
    ev = events.get(timeout=1)
    assert ev["name"]=="run_eval"
    assert ev["exec_n"]==0
    assert not ev["t1_expired"]
    assert not ev["t1_expiring"]

    # after timer t1 is expired
    ev = events.get(timeout=1)
    assert ev["name"]=="run_eval"
    assert ev["exec_n"]==1
    assert ev["t1_expired"]
    assert ev["t1_expiring"]
    # t1 will expire after 0.5s as it is reset by the second call to tmrSet 
    delta = ev["delta"].seconds + ev["delta"].microseconds/1e6
    assert delta == pytest.approx(0.5, rel=1e-2)

    # after timer t2 is expired
    ev = events.get(timeout=1)
    assert ev["name"]=="run_eval"
    assert ev["exec_n"]==2
    assert ev["t1_expired"]
    assert not ev["t1_expiring"]
    # t2 will expire after 0.7s as the reset on second call is false
    delta = ev["delta"].seconds + ev["delta"].microseconds/1e6
    assert delta == pytest.approx(0.7, rel=1e-2)

    with pytest.raises(Empty):
        ev = events.get(timeout=0.5)
